import inspect
from functools import wraps, reduce
from itertools import chain as chain_iters
from typing import Callable, Dict, Optional, List

from tcs_protowhat.State import State
from tcs_protowhat.failure import Failure, InstructorError, _debug


def state_dec_gen(sct_dict: Dict[str, Callable]):
    def state_dec(f):
        """Decorate check_* functions to return F chain if no state passed"""

        @wraps(f)
        def wrapper(*args, **kwargs):
            state = kwargs.get("state", args[0] if len(args) else None)
            if isinstance(state, State):
                return f(*args, **kwargs)
            else:
                return LazyChain(
                    ChainedCall(f, args, kwargs), chainable_functions=sct_dict
                )

        return wrapper

    return state_dec


def get_check_name(check) -> str:
    # Probe objects have a test_name
    return getattr(check, "__name__", getattr(check, "test_name", type(check).__name__))


def link_to_state(check: Callable[..., State]) -> Callable[..., State]:
    @wraps(check)
    def wrapper(state, *args, **kwargs):
        new_state = None
        error = None
        should_debug = False
        try:
            new_state = check(state, *args, **kwargs)
        except Failure as exception:
            error = exception
            # TODO: add debug information to student failure in correct environment
            # Prevent double debugging
            # - by a manual debug call
            # - by a logic function capturing an inner debug (keeping only the debug conclusion)
            should_debug = (isinstance(error, InstructorError)) and get_check_name(
                check
            ) not in ["_debug", "multi", "check_correct", "check_or", "check_not"]

            if should_debug:
                # Try creating a child state to set creator info
                # without overriding earlier creator info
                try:
                    new_state = state.to_child(error.feedback.conclusion)
                except InstructorError:
                    pass

        if not new_state:
            new_state = state

        if new_state != state and hasattr(new_state, "creator"):
            ba = inspect.signature(check).bind(state, *args, **kwargs)
            ba.apply_defaults()
            new_state.creator = {
                "type": get_check_name(check),
                "args": {**(new_state.creator or {}).get("args", {}), **ba.arguments},
            }

        if error:
            if should_debug:
                # The force flag prevents elevating a student failure with debugging info
                # to InstructorError, which would break SCTs
                _debug(
                    new_state,
                    "\n\nDebug on error:",
                    force=isinstance(error, InstructorError),
                )

            raise error

        return new_state

    return wrapper


class ChainedCall:
    strict = False
    __slots__ = ("callable", "args", "kwargs")

    def __init__(
        self,
        callable_: Callable,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ):
        """
        This contains the data for a function call that can be chained
        This means the chain state should only be provided when calling.
        """
        self.callable = link_to_state(callable_)
        self.args = args or ()
        self.kwargs = kwargs or {}
        if self.strict:
            self.validate()

    def validate(self) -> bool:
        """Can be used to check if the call data is valid without execution in the future."""
        raise NotImplementedError

    def __call__(self, state: State) -> State:
        return self.callable(state, *self.args, **self.kwargs)

    def __str__(self):
        return (
            self.callable.__name__
            + "("
            + ", ".join(
                chain_iters(
                    (str(arg) for arg in self.args),
                    (
                        "{}={}".format(kwarg, value)
                        for kwarg, value in self.kwargs.items()
                    ),
                )
            )
            + ")"
        )


class Chain:
    empty_call_str = ""

    def __init__(
        self,
        chained_call: Optional[Callable] = None,
        previous: Optional["Chain"] = None,
        chainable_functions: Dict[str, Callable] = None,
    ):
        if not chained_call and previous:
            raise ValueError("After the start of a chain a call is required")

        self.call = chained_call
        self.previous = previous
        self.next = []

        if self.previous:
            previous.next.append(self)

        if chainable_functions:
            self.chainable_functions = chainable_functions
        elif previous:
            self.chainable_functions = previous.chainable_functions
        else:
            self.chainable_functions = {}

    @property
    def _history(self) -> list:
        return getattr(self.previous, "_history", []) + [self]

    def __getattr__(self, attr):
        chainable_functions = self.chainable_functions
        if attr not in chainable_functions:
            raise AttributeError("No function named %s" % attr)
        else:
            # in case someone does: a = chain.a; a(...); a(...)
            return ChainExtender(self, chainable_functions[attr])

    def __rshift__(self, f: Callable) -> "Chain":
        if isinstance(f, EagerChain):
            raise BaseException(
                "did you use a result of the Ex() function on the right hand side of the >> operator?"
            )
        elif not callable(f):
            raise BaseException("right hand side of >> operator should be callable!")

        # wrapping the lazy chain makes it possible to reuse lazy chains
        # while still keeping a unique upstream chain (needed to lazily execute chains)
        # during execution, the state provides access to the full upstream context for an invocation
        return type(self)(f, self)

    def __call__(self, state) -> State:
        # running the chain (multiple runs possible)
        return reduce(
            lambda s, call: call(s),
            (chain.call for chain in self._history if chain.call is not None),
            state,
        )

    def __str__(self):
        return ".".join(
            str(step.call) for step in self._history if step.call is not None
        )


class LazyChain(Chain):
    """
    This is an alias for Chain
    It is useful to refer to 'pure' chains in instance checks and type hints
    """

    pass


class EagerChain(Chain):
    def __init__(
        self,
        chained_call: Optional[Callable] = None,
        previous: Optional["EagerChain"] = None,
        chainable_functions: Dict[str, Callable] = None,
        state: Optional[State] = None,
    ):
        super().__init__(chained_call, previous, chainable_functions)
        if previous is not None and state is not None:
            raise ValueError(
                "State should be set at the start. "
                "After that a reference to the previous part of the chain is needed."
            )

        if previous:
            state = previous._state

        if state and chained_call:
            self._state = chained_call(state)
        else:
            self._state = state

    def __str__(self):
        if self.call is None:
            result = "Ex()"
        else:
            result = str(self.call)

        if len(result) and len(self.next) > 0:
            result += "."

        if len(self.next) == 1:
            result += str(self.next[0])
        elif len(self.next) > 1:
            result += "multi({})".format(", ".join(map(str, self.next)))

        return result


class ChainExtender:
    """
    This handles branching off (multiple next)
    either from a Chain or a chain attribute
    """

    def __init__(self, chain: Chain, function: Callable):
        self.chain = chain
        self.function = function

    def __call__(self, *args, **kwargs) -> Chain:
        return type(self.chain)(
            ChainedCall(self.function, args, kwargs), previous=self.chain
        )

    def __getattr__(self, item):
        self.invalid_next_step(item)

    def __rshift__(self, other):
        self.invalid_next_step(other)

    def invalid_next_step(self, next_step: str):
        raise AttributeError(
            "Expected a call of {} before {}. ".format(
                getattr(self.function, "__name__", repr(self.function)), next_step
            )
        )


def get_chain_ends(chain: Chain) -> List[Chain]:
    if chain.next:
        return [*chain_iters(*(get_chain_ends(branch) for branch in chain.next))]
    else:
        return [chain]


class ChainStart:
    """Create new chains and keep track of the created chains"""

    def __init__(self, chainable_functions: Dict[str, Callable]):
        self.chain_roots = []
        # not copied on purpose to have a dynamic function registry
        self.chainable_functions = chainable_functions

    def __call__(self) -> Chain:
        """Create a new chains and store it"""
        raise NotImplementedError()

    def register_chainable_function(self, function: Callable, name: str = None):
        """This mutates the chainable_functions dictionary passed on init

        This way all chains are aware of the extra function (Ex, F and decorated)
        """
        name = name if name is not None else function.__name__
        self.chainable_functions[name] = function


class LazyChainStart(ChainStart):
    def __call__(self) -> LazyChain:
        chain_root = LazyChain(chainable_functions=self.chainable_functions)
        self.chain_roots.append(chain_root)

        return chain_root

    def __str__(self):
        return "\n".join(
            str(chain_end)
            for chain_end in chain_iters(
                *(get_chain_ends(root) for root in self.chain_roots)
            )
        )


class ExGen(ChainStart):
    def __init__(self, sct_dict: Dict[str, Callable], root_state, strict=True):
        super().__init__(sct_dict)
        self.root_state = root_state
        self.strict = strict

    def __call__(self, state=None) -> EagerChain:
        """Returns the current code state as an EagerChain instance.

        This allows SCTs to be run without including their 1st argument, ``state``.

        When writing SCTs on DataCamp, no State argument to ``Ex`` is necessary.
        The exercise State is built for you.

        Args:
            state: a State instance, which contains the student/solution code and results.

        :Example:

            code ::

                # How to write SCT on DataCamp.com
                Ex().has_code(text="SELECT id")

                # Experiment locally - chain off of Ex(), created from state
                state = SomeStateProducingFunction()
                Ex(state).has_code(text="SELECT id")

                # Experiment locally - no Ex(), create and pass state explicitly
                state = SomeStateProducingFunction()
                has_code(state, text="SELECT id")

        """
        if self.strict and state is None and self.root_state is None:
            raise Exception("explicitly pass state to Ex, or set Ex.root_state")

        chain_root = EagerChain(
            chainable_functions=self.chainable_functions,
            state=state or self.root_state,
        )
        self.chain_roots.append(chain_root)

        return chain_root

    def __str__(self):
        return "\n".join(str(root) for root in self.chain_roots)



import builtins

from functools import wraps
from importlib import import_module
from typing import Dict, Callable, Type, Any, Tuple

from tcs_protowhat.Reporter import Reporter
from tcs_protowhat.State import State
from tcs_protowhat.sct_syntax import (
    ExGen,
    LazyChainStart,
    EagerChain,
    state_dec_gen,
)
from tcs_protowhat.utils import get_class_parameters


def get_checks_dict(checks_module) -> Dict[str, Callable]:
    return {
        k: v
        for k, v in vars(checks_module).items()
        if k not in builtins.__dict__
        if not k.startswith("__")
        if callable(v)
    }


def create_sct_context(sct_dict, root_state: State = None) -> Dict[str, Callable]:
    """
    Create the globals that will be available when running the SCT.

    Args:
        sct_dict: a dictionary of the functions to make available
        root_state: a State instance with the exercise information available to the SCT

    Returns:
        dict: the globals available to the SCT code
    """
    state_dec = state_dec_gen(sct_dict)
    sct_ctx = {k: state_dec(v) for k, v in sct_dict.items()}

    ctx = {
        **sct_ctx,
        "state_dec": state_dec,  # needed by ext packages
        "Ex": ExGen(sct_dict, root_state),
        "F": LazyChainStart(sct_dict),
    }

    return ctx


def create_embed_state(
    xstate: Type[State],
    parent_state: State,
    derive_custom_state_args: Callable[[State], Dict[str, Any]] = None,
) -> State:
    """
    Create the state for running checks in the embedded technology.

    This function also connects the created state with the state of the host technology.

    Args:
        xstate: the State class of the embedded technology
        parent_state: state of the host technology to derive the embedded state from
        derive_custom_state_args: function to calculate instructor extra arguments to pass to the constructor of the embedded state

    Returns:
        an instance of xstate
    """
    # gather the kwargs the xstate will be created with
    args = {}

    # find all arguments the xstate constructor can handle
    embedded_state_parameters = getattr(
        xstate, "parameters", get_class_parameters(xstate)
    )

    # copy all allowed arguments from the parent state
    for arg in embedded_state_parameters:
        if hasattr(parent_state, arg):
            args[arg] = getattr(parent_state, arg)

    # configure the reporter to collaborate with the parent state reporter
    args["reporter"] = Reporter(parent_state.reporter)

    custom_args = (
        derive_custom_state_args(parent_state) if derive_custom_state_args else {}
    )

    # manually add / override arguments
    args.update(**custom_args)

    embed_state = xstate(**args)
    # TODO: other params? set manually through chain constructor or add State args
    # to pass: path, debug; don't pass: highlight, ast_dispatcher, params
    embed_state.creator = {"type": "embed", "args": {"state": parent_state}}

    return embed_state


def create_embed_context(
    technology: str, context: EagerChain, derive_custom_state_args=None
):
    """
    Create the globals that will be available when running the checks for the embedded technology.

    Args:
        technology: the name of the embedded technology (the x in xwhat)
        context: the Chain of the host technology
            the checks for the embedded technology will use as starting point
        derive_custom_state_args: passed to create_embed_state

    Returns:
        dict: the variables available to the SCT code for the embedded technology
    """
    parent_state = context._state

    xwhat = import_module("{}what".format(technology))
    xwhat_checks = import_module("{}what.checks".format(technology))
    xwhat_state = xwhat.State.State

    xwhat_scts = get_checks_dict(xwhat_checks)

    embed_state = create_embed_state(
        xwhat_state, parent_state, derive_custom_state_args
    )

    return create_sct_context(xwhat_scts, root_state=embed_state)


def get_embed_chain_constructors(*args, **kwargs) -> Tuple[Type, Type]:
    """
    Get the chain constructors for the embedded technology.

    This is a wrapper around create_embed_context

    Returns:
        tuple: Ex and F for the embedded xwhat
    """
    new_context = create_embed_context(*args, **kwargs)
    return new_context["Ex"], new_context["F"]



from contextlib import contextmanager
from typing import TYPE_CHECKING, List, Generator, Callable

from tcs_protowhat.Feedback import Feedback, FeedbackComponent

if TYPE_CHECKING:
    from tcs_protowhat.State import State


def check_history(state_history: List["State"]) -> Generator[str, None, None]:
    return (state.creator["type"] for state in state_history if state.creator)


def invert_failure(state: "State") -> bool:
    return "check_not" in check_history(state.state_history)


class Failure(Exception):
    throwing = False

    def __init__(self, feedback: Feedback, state_history: List["State"]):
        if not isinstance(feedback, Feedback):
            raise ValueError("Use the from_message method")
        super().__init__(feedback)
        self.feedback = feedback
        self.state_history = state_history

    def __str__(self):
        # get_message can be expensive
        # TODO: check speed
        return self.feedback.get_message()

    @classmethod
    def from_message(cls, message: str) -> "Failure":
        return cls(Feedback(FeedbackComponent(message)), [])


class TestFail(Failure):
    pass


class InstructorError(Failure):
    pass


@contextmanager
def debugger(state: "State", allow_failure: Callable[["State"], bool] = invert_failure):
    # TODO: skip debugging in production
    debugging = state.debug
    skippable = False  # not state.force_diagnose
    state.debug = True
    try:
        yield skippable
    except InstructorError as e:
        if not allow_failure(state):
            raise e

    state.debug = debugging


def _debug(state: "State", msg="", on_error=False, force=True):
    """
    This SCT function makes the SCT fail with a message containing debugging information
    and highlights the focus of the SCT at that point.

    To make the interruption behave like a student failure, use ``force=False``.
    """
    checks = check_history(state.state_history)

    feedback = ""
    if msg:
        feedback += msg + "\n"

    if checks:
        feedback += "SCT function state history: `{}`".format(" > ".join(checks))

    if state.reporter.tests:
        feedback += "\nLast test: `{}`".format(repr(state.reporter.tests[-1]))

    if on_error:
        # debug on next failure
        state.debug = True
        # or at the end (to prevent debug mode in production)
        state.reporter.fail = True
    else:
        # latest highlight added automatically
        failure_type = InstructorError if force else TestFail
        raise failure_type(state.get_feedback(FeedbackComponent(feedback)), state.state_history)

    return state



import itertools
from functools import wraps
from inspect import signature, Parameter
from typing import Type, Iterator, Callable, Dict


def get_class_parameters(cls: Type) -> Iterator[str]:
    """
    Get an iterator over all arguments that can be used to construct a class.

    This works for class hierarchies that have compatible constructors.

    Args:
        cls: class to inspect constructor parameters in class hierarchy for

    Returns:
        Iterator over parameter names
    """
    return itertools.chain(
        *(get_class_parameters(cls) for cls in cls.__bases__),
        (
            name
            for name, param in signature(cls).parameters.items()
            if param.kind is not Parameter.VAR_POSITIONAL
            and param.kind is not Parameter.VAR_KEYWORD
        )
    )


def parameters_attr(cls: Type) -> Type:
    """Add a parameters attribute to the class with all arguments that can be used to construct it.

    Why?
    - calculated once: not manual list, not built every init
    - accessible on class: not only usable on instances
    - set difference between vars() and parameters doesn't include this class attribute
    - more noticeable than manually setting attribute after class definition

    Args:
        cls: class to add parameters attribute to

    Returns:
        cls: modified class
    """
    cls.parameters = list(get_class_parameters(cls))
    return cls


def legacy_signature(
    **kwargs_mapping: Dict[str, str]
) -> Callable[[Callable], Callable]:
    """
    This decorator makes it possible to call a function using old argument names
    when they are passed as keyword arguments.

    :Example:

        @legacy_signature(old_arg1='arg1', old_arg2='arg2')
        def func(arg1, arg2=1):
            return arg1 + arg2

        func(old_arg1=1) == 2
        func(old_arg1=1, old_arg2=2) == 3
    """

    def signature_decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            redirected_kwargs = {
                kwargs_mapping[k] if k in kwargs_mapping else k: v
                for k, v in kwargs.items()
            }
            return f(*args, **redirected_kwargs)

        return wrapper

    return signature_decorator



from typing import Dict, Union, List
from collections import Counter
from jinja2 import Template


class FeedbackComponent:
    def __init__(self, message: str, kwargs: dict = None, append=True):
        self.message = message
        self.kwargs = kwargs or {}
        self.append = append

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, repr(vars(self)))


class Feedback:
    # This is the offset for ANTLR
    ast_highlight_offset = {"column_start": 1, "column_end": 1}

    def __init__(
        self,
        conclusion: FeedbackComponent,
        context_components: List[FeedbackComponent] = None,
        highlight=None,
        path=None,
        highlighting_disabled=False,
        highlight_offset=None,
        full_code_position=None,
    ):
        self.conclusion = conclusion
        self.context_components = context_components or []
        self.highlight = highlight
        self.path = path
        self.highlighting_disabled = highlighting_disabled
        self.highlight_offset = highlight_offset or {}
        self.full_code_position = full_code_position

    @classmethod
    def get_highlight_position(cls, highlight) -> Dict[str, int]:
        """Get the position of the highlight object

        This method can be customized in subclasses and
        use class properties to do so if needed
        """
        if hasattr(highlight, "get_position"):
            return highlight.get_position()

    def get_highlight(self) -> Dict[str, int]:
        highlight = Counter()

        if not self.highlighting_disabled:
            position = self.get_highlight_position(self.highlight)

            # TODO: handle highlighting everything
            # if position != self.full_code_position:

            if position:
                highlight.update(self.highlight_offset)
                if "line_start" in highlight and "line_end" not in highlight:
                    highlight["line_end"] = highlight["line_start"]

                highlight.update(position)
                highlight.update(self.ast_highlight_offset)

        if self.path:
            highlight["path"] = str(self.path)

        return highlight or {}

    def get_message(self):
        out_list = []
        msgs = [*filter(lambda x: x is not None, self.context_components), self.conclusion]

        if not self.conclusion.append:
            msgs = msgs[-1:]
        elif getattr(self, "highlight", None) and not self.highlighting_disabled:
            # if highlighting info is available, don't use all context messages
            msgs = msgs[-3:]

        # format messages in list, by iterating over previous, current, and next message
        for prev_msg, msg, next_msg in zip([None, *msgs[:-1]], msgs, [*msgs[1:], None]):
            tmp_kwargs = {
                "parent": getattr(prev_msg, "kwargs", None),
                "child": getattr(next_msg, "kwargs", None),
                "this": getattr(msg, "kwargs"),
                **getattr(msg, "kwargs"),
            }
            # don't bother appending if there is no message
            if not getattr(msg, "message"):
                continue
            else:
                # This is slow (but it should only happen once for a submission)
                out = Template(msg.message.replace("__JINJA__:", "")).render(tmp_kwargs)
                out_list.append(out)

        return "".join(out_list)

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, repr(vars(self)))



from copy import copy
from typing import Union

from tcs_protowhat.selectors import DispatcherInterface
from tcs_protowhat.Feedback import Feedback, FeedbackComponent
from tcs_protowhat.Test import Fail, Test
from tcs_protowhat.failure import TestFail, debugger, InstructorError
from tcs_protowhat.utils import parameters_attr


class DummyDispatcher(DispatcherInterface):
    def __init__(self):
        class ParseError(Exception):
            pass

        self.ParseError = ParseError

    def find(self, name, node, *args, **kwargs):
        return []

    def select(self, path: str, node):
        return None

    def parse(self, code):
        return self.ParseError()

    def describe(self, *args, **kwargs):
        return "code"


@parameters_attr
class State:
    feedback_cls = Feedback

    def __init__(
        self,
        student_code,
        solution_code,
        pre_exercise_code,
        student_conn,
        solution_conn,
        student_result,
        solution_result,
        reporter,
        force_diagnose=False,
        highlight_offset=None,
        highlighting_disabled=False,
        feedback_context=None,
        creator=None,
        solution_ast=None,
        student_ast=None,
        ast_dispatcher=None,
    ):
        args = locals().copy()
        self.debug = False

        for k, v in args.items():
            if k != "self":
                setattr(self, k, v)

        if ast_dispatcher is None:
            self.ast_dispatcher = self.get_dispatcher()

        # Parse solution and student code
        # if possible, not done yet and wanted (ast arguments not False)
        if isinstance(self.solution_code, str) and self.solution_ast is None:
            with debugger(self):
                self.solution_ast = self.parse(self.solution_code)
        if isinstance(self.student_code, str) and self.student_ast is None:
            self.student_ast = self.parse(self.student_code)

    def parse(self, text):
        result = None
        if self.ast_dispatcher:
            try:
                result = self.ast_dispatcher.parse(text)
            except self.ast_dispatcher.ParseError as e:
                if self.debug:
                    self.report(
                        "Something went wrong when parsing PEC or solution code: %s"
                        % str(e)
                    )
                else:
                    self.report(e.message)

        return result

    def get_dispatcher(self):
        return DummyDispatcher()

    @property
    def parent_state(self):
        if self.creator is not None:
            creator_args = self.creator.get("args")
            if creator_args:
                return creator_args.get("state")

    @property
    def is_root(self):
        return self.parent_state is None

    @property
    def state_history(self):
        return getattr(self.parent_state, "state_history", []) + [self]

    def get_ast_path(self):
        rev_checks = filter(
            lambda x: x.creator is not None
            and x.creator["type"] in ["check_edge", "check_node"],
            reversed(self.state_history),
        )
        try:
            last = next(rev_checks)
            if last.creator["type"] == "check_node":
                # final check was for a node
                return self.ast_dispatcher.describe(
                    last.student_ast,
                    index=last.creator["args"]["index"],
                    msg="{index}{node_name}",
                )
            else:
                node = next(rev_checks)
                if node.creator["type"] == "check_node":
                    # checked for node, then for target, so can give rich description
                    return self.ast_dispatcher.describe(
                        node.student_ast,
                        field=last.creator["args"]["name"],
                        index=last.creator["args"]["index"],
                        msg="{index}{field_name} of the {node_name}",
                    )

        except StopIteration:
            return self.ast_dispatcher.describe(self.student_ast, "{node_name}")

    def report(self, feedback: str, kwargs=None, append=True):
        test_feedback = FeedbackComponent(feedback, kwargs, append)
        test = Fail(test_feedback)

        return self.do_test(test)

    def do_test(self, test: Test):
        result, test_feedback = self.reporter.do_test(test)
        if result is False:
            failure_type = InstructorError if self.debug else TestFail
            raise failure_type(self.get_feedback(test_feedback), self.state_history)
        return result, test_feedback

    def do_tests(self, tests):
        return [self.do_test(test) for test in tests]

    def get_feedback(self, conclusion):
        full_code_position = self.feedback_cls.get_highlight_position(
            self.state_history[0].student_ast
        )

        return self.feedback_cls(
            conclusion,
            [state.feedback_context for state in self.state_history],
            getattr(self, "highlight", self.student_ast),
            getattr(self, "path", None),
            self.highlighting_disabled,
            self.highlight_offset,
            full_code_position,
        )

    def to_child(self, append_message: Union[str, FeedbackComponent] = None, **kwargs):
        """Basic implementation of returning a child state"""

        bad_parameters = set(kwargs) - set(self.parameters)
        if bad_parameters:
            raise ValueError(
                "Invalid init parameters for State: %s" % ", ".join(bad_parameters)
            )

        if append_message and not isinstance(append_message, FeedbackComponent):
            if isinstance(append_message, str):
                append_message = FeedbackComponent(append_message)
            else:
                raise ValueError(
                    "append_message should be a FeedbackComponent or a string"
                )
        kwargs["feedback_context"] = append_message
        kwargs["creator"] = {"type": "to_child", "args": {"state": self}}

        child = copy(self)
        for k, v in kwargs.items():
            setattr(child, k, v)

        return child
